#include <vector>
#include <cstdio>

#include "atoms.hh"
#include "level.hh"
#include "pairs.hh"
#include "handling.hh"
#include "rom.hh"
#include "objects.hh"

static void LoadScreen(Screen& target,
                       unsigned roomid,
                       int cont_dir)
{
    const Room& room = Rooms[roomid];
    target.cont_dir = cont_dir;
    
    for(unsigned y=0; y<16; ++y)
        for(unsigned x=0; x<16; ++x)
        {
            unsigned blockid = room.blocks[(x/2)*8 + (y/2)];
            const Block& b = Blocks[blockid];
            
            target.outmap[y*16+x] = b.c[(x%2)*2 + (y%2)];
        }
}
                       
static unsigned LoadLevel(Level& target, unsigned first_order)
{
    unsigned first_screen    = ROM.Read(Addr_FirstScreenTable + target.LevelNumber);
    unsigned restart1_screen = ROM.Read(Addr_RestartPointsScreenTable1 + target.LevelNumber);
    unsigned restart2_screen = ROM.Read(Addr_RestartPointsScreenTable2 + target.LevelNumber);
    
    if(target.LevelNumber == 11)
    {
        first_screen = ROM.Read(0x06AB11);
        /* last_screen = ROM.Read(0x06AB15); */
    }

    unsigned screens_addr = BaseAddr + Addr_Screens + first_screen;
    
    for(unsigned a=0; a<MaxRooms; ++a)
        target.rooms[a].cont_dir = NoRooms;

    unsigned screen    = first_screen;
    unsigned cur_order = first_order;
    
    for(;;)
    {
        const unsigned order_addr = BaseAddr + Addr_Layout + cur_order++;
        unsigned char b = ROM.Read(order_addr);
        
        unsigned n_screens = (b & 0x1F);
        for(unsigned a=0; a<=n_screens; ++a, ++screen)
        {
        redo:
            const unsigned screen_addr = screens_addr++;
            unsigned char c = ROM.Read(screen_addr);
            printf("// Order=%02X(@%04X = %02X, Room(@%04X = %02X), Screen=%02X(counter=%u)\n",
                cur_order, order_addr, b,
                           screen_addr, c, // index to room gfx table
                screen, a                  // counter in this program
                  );

            if(c >= Rooms.size())
            {
                printf("c=%02X, ignoring\n", c);
                goto redo;
            }
            
            int dir = Right;
            
            if(a == 0)
            {
                if(screen == first_screen) dir |= LevelBegin;
                if(screen == restart1_screen) dir |= RestartPointA;
                if(screen == restart2_screen) dir |= RestartPointB;
            }
            
            if(a == n_screens)
            {
                if(b & 0x20) dir |= Shutter;
                if(b & 0x40) dir |= Down;
                if(b & 0x80) dir |= Up;
            }
            
            if(screen > MaxRooms)
            {
                printf("Too many rooms (%u > %u)\n",
                    screen, MaxRooms);
                continue;
            }

            LoadScreen(target.rooms[screen-first_screen], c, dir);
            target.CreateAscii(target.rooms[screen-first_screen]);
        }
        if(!(b & 0xE0)) break;
    }
    
    return cur_order - first_order;
}

static void PutTeleportCoordinatesIn(Level& l)
{
    for(unsigned a=0; a<MaxRooms; ++a)
    {
        int y = -1;
        
        Screen& s = l.rooms[a];
        if(s.cont_dir == NoRooms) break;
        if(s.cont_dir & LevelBegin)
            y = ROM.Read(Addr_TeleportYcoordTable + l.LevelNumber) >> 4;
        if(s.cont_dir & RestartPointA)
            y = ROM.Read(Addr_RestartTeleportYcoordTable1 + l.LevelNumber) >> 4;
        if(s.cont_dir & RestartPointB)
            y = ROM.Read(Addr_RestartTeleportYcoordTable2 + l.LevelNumber) >> 4;
        
        if(y >= 0 && y <= 15)
        {
            unsigned x=8;
            if(s.ascii[y*32 + 16+x] != ' ')
            {
                for(x=0; ; ++x)
                {
                    if(x >= 16) { x=8; break; }
                    
                    if(s.ascii[y*32 + 16+x] == ' ')break;
                }
            }
            s.ascii[y*32 + 16+x] = Ascii_TeleportYcoord;
        }
    }
    if(l.LevelNumber < 8)
    {
        for(unsigned a=0; a<MaxRooms; ++a)
        {
            Screen& s = l.rooms[a];
            if(l.rooms[a+1].cont_dir == NoRooms)
            {
                unsigned bossx = ROM.Read(Addr_BossXcoord + l.LevelNumber) >> 4;
                unsigned bossy = ROM.Read(Addr_BossYcoord + l.LevelNumber) >> 4;
                s.ascii[bossy*32 + 16 + bossx] = Ascii_Boss;
                break;
            }
        }
    }
}

static void DumpLevel(const Level& l)
{
    unsigned levnum = l.LevelNumber;
    printf("Level %s =\n{\n", l.Name);
    printf("\t%u, // level number\n", levnum);
    printf("\t\"%s\",\n", l.Name);
    printf("\n\t// screens:\n\t{\n");
    for(unsigned a=0; a<MaxRooms; ++a)
    {
        if(l.rooms[a].cont_dir == NoRooms) break;
        
        printf("\t\t{\n");
        for(unsigned y=0; y<16; ++y)
        {
            printf("\t\t\"");
            for(unsigned x=0; x<32; ++x)
            {
                unsigned char asc = l.rooms[a].ascii[y*32+x];
                putchar(asc);
                if(x==15) printf("\"   \"");
            }
            printf("\"");

#if 1
            printf("   //");
            for(unsigned x=0; x<16; ++x)
            {
                const Char& chr  = l.rooms[a].outmap[y*16+x];
                printf("%u%02X-", chr.palette, chr.tile);
            }
#endif
            printf("\n");
        }
        printf("\t\t, %u // monster graphics index\n", l.rooms[a].MonsterGraphicsIndex);
        printf("\t\t,");
        const char* sep = "";
        bool stop = false;
        
        if(l.rooms[a].cont_dir & Up) { printf("%sUp",sep); sep="| "; stop=true; }
        if(l.rooms[a].cont_dir & Down) { printf("%sDown", sep); sep="| "; stop=true; }
        if(l.rooms[a].cont_dir & Shutter) { printf("%sShutter", sep); sep="| "; stop=true; }
        if(l.rooms[a].cont_dir & LevelBegin) { printf("%sLevelBegin", sep); sep="| "; }
        if(l.rooms[a].cont_dir & RestartPointA) { printf("%sRestartPointA", sep); sep="| "; }
        if(l.rooms[a].cont_dir & RestartPointB) { printf("%sRestartPointB", sep); sep="| "; }
        
        if(!stop)
        {
            bool last = (a+1)>=MaxRooms || l.rooms[a+1].cont_dir == NoRooms;
            printf("%s%s", sep, last ? "End": "Right");
        }
        
        printf("},\n");
    }
    printf("\t},\n\n");

    printf("\t// palette\n\t{\n");
    for(unsigned a=0; a<16*3; ++a)
    {
        if((a&3)==0) printf("\t\t");
        printf("%2u, ", l.Palette[a]);
        if((a&3)==3) printf("\n");
        if((a&15)==15 && a+1<16*3) printf("\n"); //group by 16
    }
    printf("\t},\n\n");

    if(levnum < 6)
    {
        printf("\t// after doors palette\n\t{\n");
        for(unsigned a=0; a<16; ++a)
        {
            if((a&3)==0) printf("\t\t");
            printf("%2u, ", l.AfterDoorPalette[a]);
            if((a&3)==3) printf("\n");
            if((a&15)==15 && a+1<16*3) printf("\n"); //group by 16
        }
        printf("\t},\n\n");
    }
    else
    {
        printf("\t// after doors palette not used in wily stages\n");
        printf("\t{},\n\n");
    }

    printf("\t// actives\n\t{\n");
    
    for(unsigned a=0; a<MaxActives; ++a)
    {
        if(l.actives[a].screen == 255) continue;
        printf("\t\tActive( %u, %u, 0x%02X,0x%02X, 0x%02X,0x%02X ),\n",
            l.actives[a].type,
            l.actives[a].screen,
            l.actives[a].x1,
            l.actives[a].y1,
            l.actives[a].x2,
            l.actives[a].y2);
    }
    printf("\t},\n\n");

    printf("\t// enemies\n\t{\n");

    for(unsigned a=0; a<MaxEnemies; ++a)
    {
        if(l.enemies[a].screen == 255) continue;
        printf("\t\tEnemy( %u, 0x%02X,0x%02X, 0x%X ),\n",
            l.enemies[a].screen,
            l.enemies[a].x,
            l.enemies[a].y,
            l.enemies[a].type);
    }
    printf("\t},\n\n");
    
    printf("\t// level gfx (id, count) (max %u); id is block*4+bank\n\t{\n\t\t", MaxLevelGfx);
    for(unsigned a=0; a<MaxLevelGfx; ++a)
    {
        if(!l.levelgfx[a].count) break;
        if(a>0 && (a%4)==0) printf("\n\t\t");
        printf("{ 0x%02X, %u }, ", l.levelgfx[a].index, l.levelgfx[a].count);
    }
    printf("\n\t},\n\n");
    
    printf("\t// monster gfx tables (max %u) (5 tiles, 2*3bytepalette)\n\t{\n", MaxMonsterGfx);
    for(unsigned a=0; a<MaxMonsterGfx; ++a)
    {
        if(!l.monstergfx[a].index[0] && !l.monstergfx[a].index[1]) break;
        printf("\t\t{ { ");
        for(unsigned b=0; b<5; ++b) printf("0x%02X, ", l.monstergfx[a].index[b]);
        printf("}, { ");
        for(unsigned b=0; b<6; ++b) printf("%u,%s", l.monstergfx[a].palette[b], b==2?" ":"");
        printf("} },\n");
    }
    printf("\t},\n\n");
    
    printf("\t0x%02X, // Arrival input\n\n",
        ROM.Read(Addr_BossArrivalInputTable + l.LevelNumber));
    
    printf("\t%sAscii, %sChar\n", l.Name, l.Name);
    printf("};\n\n\n");
}

static void LoadPair(LevelPairData& pair)
{
    BaseAddr      = pair.BeginAddr;
    BlockBaseAddr = pair.BeginAddr;

    Rooms.Load();
    Blocks.Load();
    
    unsigned cur_order=1;
    if(true)
    {
        const unsigned first_screen = ROM.Read(Addr_FirstScreenTable + pair.level1.LevelNumber);
        
        unsigned num_orders = LoadLevel(pair.level1, cur_order);
        cur_order += num_orders;
        
        pair.level1.ReadPalette(BaseAddr + Addr_Palette1,
                                Addr_AfterDoorPalette + pair.level1.LevelNumber*16);
        pair.level1.ReadActives(first_screen, BaseAddr + Addr_Actives1);
        pair.level1.ReadEnemies(first_screen, Addr_EnemyPointers + pair.level1.LevelNumber*2);
        
        PutTeleportCoordinatesIn(pair.level1);
        
        pair.level1.ReadGFX(BaseAddr + Addr_Tiles1,
                            BaseAddr + Addr_MonsterGraphicsIndex,
                            BaseAddr + Addr_MonsterGraphicsTable);

        pair.level1.AnalyzeObjects();
        
        DumpLevel(pair.level1);
    }
    
    //if(&pair.level1 != &pair.level2)
    {
        const unsigned first_screen = ROM.Read(Addr_FirstScreenTable + pair.level2.LevelNumber);
        
        unsigned num_orders = LoadLevel(pair.level2, cur_order);
        //cur_order += num_orders;
        
        pair.level2.ReadPalette(BaseAddr + Addr_Palette2, 
                                Addr_AfterDoorPalette + pair.level2.LevelNumber*16);

        pair.level2.ReadActives(first_screen, BaseAddr + Addr_Actives2);
        pair.level2.ReadEnemies(first_screen, Addr_EnemyPointers + pair.level2.LevelNumber*2);

        PutTeleportCoordinatesIn(pair.level2);
        
        pair.level2.ReadGFX(BaseAddr + Addr_Tiles2,
                            BaseAddr + Addr_MonsterGraphicsIndex + first_screen,
                            BaseAddr + Addr_MonsterGraphicsTable);

        pair.level2.AnalyzeObjects();
    
        DumpLevel(pair.level2);
    }
}

int main(int argc, const char*const* argv)
{
    if(argv[1])
    {
        InputFN = argv[1];
        ROM.ReRead();
    }
    OutputFN = "/dev/null";
    
    LoadPair(pair5);/*
    LoadPair(pair1);
    LoadPair(pair2);
    LoadPair(pair3);
    LoadPair(pair4);
    LoadPair(pair5);
    LoadPair(pair6);*/

    return 0;
}
