#include "rom.hh"
#include "defs.hh"

void NopRegion(unsigned address, unsigned bytes)
{
    unsigned remain = bytes;
    unsigned char replacement[bytes];

    unsigned target = 0;

    /* If the range is small, we create JMPs. */
    while(remain > 3)
    {
        unsigned endpos = address+target+remain;
        replacement[target++] = 0x4C; // JMP
        replacement[target++] = (endpos   ) &0xFF;
        replacement[target++] = (endpos>>8) &0xFF;
        remain -= 3;
        
        unsigned freepos = address+target;
        
        fprintf(stderr, "Adding free(%06X, %u)\n", address+target, endpos-freepos);
        
        ROM.AddFreeSpace(address+target, endpos-freepos);
        remain = address+bytes - endpos;
    }

    while(target < bytes) { replacement[target++] = 0xEA; }
    ROM.WriteBytes(address, bytes, replacement);
}

void ExpandBlockMaps()
{
    // Remap Wily stages. The bank was (stage-6), it's now (stage+1).
    ROM.Write(0xC3A7, 0xFF);
    // sbc -1 is equal to adc +1, and carry was set so "sbc -1" does "+1" nicely.
    
    // Overwrite to banks 7,8,9,A,B
    // F = last page
    // E = for custom code, should it be needed.
    // D = code from custom A65/IPS files...
    for(unsigned bank=0; bank<5; ++bank)
    {
        //if(bank == 4) continue;
        
        unsigned orig_addr = MakeNESaddr(bank,   0x0000);
        unsigned new_addr  = MakeNESaddr(7+bank, 0x0000);
        
        unsigned char Buf[0x1000];
        ROM.ReadData(orig_addr, 0x1000, Buf);
        ROM.WriteBytes(new_addr,0x1000, Buf);
        
        ROM.DelFreeSpace(new_addr,0x1000);
    }
    
    /* The purpose of "squeezes" is to gain more room 
     * for the "block" array (different 4-tile squares)
     */
    
    unsigned squeeze = 0;
    
    #define Update(var) \
        fprintf(stderr, "Was %04X, redefined as %04X: %s\n", var,var+squeeze, #var); \
        var += squeeze;
    #define Squeeze(var1,var2, end) \
        fprintf(stderr, "Combining %s to %s gives %04X bytes of space\n", \
            #var2,#var1, (end-squeeze-var2)); \
        squeeze += (end-squeeze - var2); \
        Update(var1); var2 = var1;
    
    fprintf(stderr, "Before squeezes, there is room for %u blocks\n",
        Addr_BlockPals/4);

/**/
    /* Squeeze the region normally used for nothing but
     * Nintendo USA copyright licensed blah blah messages.
     */
    squeeze += 0x40; Update(Addr_ActivesEnd);
    
    Squeeze(Addr_Actives1, Addr_Actives2, Addr_ActivesEnd); // Squeezed 0x60 bytes.
    Update(Addr_MonsterGraphicsTable);
    Update(Addr_MonsterGraphicsIndex);
    Squeeze(Addr_Tiles1, Addr_Tiles2, Addr_MonsterGraphicsIndex); // Squeezed 0x20 bytes.
    Squeeze(Addr_Palette1, Addr_Palette2, Addr_Tiles1); // Squeezed 0x30 byts.
    Update(Addr_Layout);
    Update(Addr_Pointers);
    Update(Addr_Screens);
    Addr_Rooms += 64; // 1 room less. (was 33, should be 32 since there are 32 pointers)
    Update(Addr_Rooms);
/**/
    Addr_BlockPals = (Addr_Rooms/5)*4;
    
    fprintf(stderr, "After squeezes, there is now room for %u blocks\n",
        Addr_BlockPals/4);
    
    ROM.WriteWord(0xC764 +1, 0x8000 + Addr_Tiles2);
    ROM.WriteWord(0xC769 +1, 0x8000 + Addr_Tiles1);
    ROM.WriteWord(0xC783 +1, 0x8000 + Addr_Tiles2+1);
    ROM.WriteWord(0xC786 +1, 0x8000 + Addr_Tiles2+2);
    ROM.WriteWord(0xC78B +1, 0x8000 + Addr_Tiles1+2);
    ROM.WriteWord(0xC78E +1, 0x8000 + Addr_Tiles1+1);
    ROM.Write(0xC48C +1,     (0x8000 + Addr_Palette2) & 0xFF);
    ROM.Write(0xC490 +1,     (0x8000 + Addr_Palette2) >> 8);
    ROM.Write(0xC497 +1,     (0x8000 + Addr_Palette1) & 0xFF);
    ROM.Write(0xC49B +1,     (0x8000 + Addr_Palette1) >> 8);
    ROM.WriteWord(0xC61A +1, (0x8000 + Addr_Layout) + 1);
    
    ROM.WriteWord(0xCCDF +1, (0x8000 + Addr_Pointers) + 0);
    ROM.WriteWord(0xCCE4 +1, (0x8000 + Addr_Pointers) + 1);
    ROM.WriteWord(0xCE5F +1, (0x8000 + Addr_Pointers) + 0);
    ROM.WriteWord(0xCE64 +1, (0x8000 + Addr_Pointers) + 1);
    ROM.WriteWord(0xCF8E +1, (0x8000 + Addr_Pointers) + 0);
    ROM.WriteWord(0xCF93 +1, (0x8000 + Addr_Pointers) + 1);
    
    ROM.WriteWord(0xCCDA +1, (0x8000 + Addr_Screens));
    ROM.WriteWord(0xCE5A +1, (0x8000 + Addr_Screens));
    ROM.WriteWord(0xCF89 +1, (0x8000 + Addr_Screens));
    
    ROM.WriteWord(0xCE8E +1, (0x8000 + Addr_BlockPals));
    ROM.WriteWord(0xD03B +1, (0x8000 + Addr_BlockPals));
    
    ROM.Write(0xC4B4 +1,     (0x8000 + Addr_Actives2-1) & 0xFF);
    ROM.Write(0xC4B8 +1,     (0x8000 + Addr_Actives2-1) >> 8);
    ROM.Write(0xC4BE +1,     (0x8000 + Addr_Actives1-1) & 0xFF);
    ROM.Write(0xC4C2 +1,     (0x8000 + Addr_Actives1-1) >> 8);

    ROM.WriteWord(0xC6EE +1, (0x8000 + Addr_MonsterGraphicsTable));
    ROM.WriteWord(0xC711 +1, (0x8000 + Addr_MonsterGraphicsTable));
    ROM.WriteWord(0xC65E +1, (0x8000 + Addr_MonsterGraphicsIndex));
    
    
    {
        unsigned char Buf[0x1000];
        unsigned char NewPage[0x1000] = {0};
        
        /* Move the intro screen data to safer location */
        ROM.ReadData(  0x0B8EA1, 351, Buf);
        std::memcpy(NewPage+0x0000, Buf, 351);

        /* Then remap a few things */
        ROM.ReadData(  0x0B8D20,               32, Buf);
        std::memcpy(NewPage+Addr_Tiles2,  Buf, 32);
        
        ROM.ReadData(  0x0B8EA0,                 7,Buf);
        std::memcpy(NewPage+Addr_Actives2, Buf, 7);

        ROM.ReadData(  0x0B8CD0,                 32,Buf);
        std::memcpy(NewPage+Addr_Palette2, Buf, 32);
        
        /* Instead of writing directly to the ROM, we write to this
         * preblanked buffer so that the page $0B will actually only
         * contain the data that was needed, and nothing else.
         */
        ROM.WriteBytes(0x0B8000, 0x1000, NewPage);
        
        /* Patch the ROM so that it knows how to load the intro screen */
        #define DecP(addr) ROM.WriteWord(addr, ROM.ReadWord(addr) - 0xEA1)
        
        DecP(0xF983 +1);
        DecP(0xF991 +1);
        DecP(0xF997 +1);
        DecP(0xF99F +1);
        //
        DecP(0xF9D7 +1);
        DecP(0xF9DE +1);
        DecP(0xF9E5 +1);
        DecP(0xF9E9 +1);
        //
        DecP(0xFA73 +1);
        DecP(0xFA7A +1);
        DecP(0xFA81 +1);
        DecP(0xFA85 +1);
        
        #undef DecP
    }
    
    if(1){
    FILE*fp = fopen("startup.ips", "rb");
    if(fp)
    {
        ROM.LoadIPSfile(fp, "startup");
        fclose(fp);
    }}
}

void RemoveDelays()
{
    /* Mark the end of BANK 5 as free. */
    ROM.AddFreeSpace(0x05BF76, 0xC000-0xBF76);

    if(1){FILE*fp = fopen("removedelays.ips", "rb");
    if(fp)
    {
        ROM.LoadIPSfile(fp, "removedelays");
        fclose(fp);
    }}
    if(1){FILE*fp = fopen("music.ips", "rb");
    if(fp)
    {
        ROM.LoadIPSfile(fp, "music");
        fclose(fp);
    }}
    if(1){FILE*fp = fopen("shutters.ips", "rb");
    if(fp)
    {
        ROM.LoadIPSfile(fp, "shutters");
        fclose(fp);
    }}
    if(1){FILE*fp = fopen("tempblocks.ips", "rb");
    if(fp)
    {
        ROM.LoadIPSfile(fp, "tempblocks");
        fclose(fp);
    }}
    
#if 0 /* Make vertical scrolling faster. DOESN'T WORK YET */
    unsigned char NMICopyBuf[512];
    
    unsigned NMIoffs=0;
    #define Copy(begin,end) \
      ROM.ReadData(begin, end-begin, NMICopyBuf+NMIoffs); NMIoffs+=(end-begin);
     
    Copy(0xD4B7, 0xD4C7);
    Copy(0xD4CE, 0xD4F6);
    Copy(0xD52F, 0xD54A);
    ROM.Write(NMIoffs++, 0x60); // RTS
    #undef Copy
    
    unsigned AddrNMICopyBuf = ROM.FindFreeSpaceIn(ROMmap_npages-1, NMIoffs);
    ROM.WriteBytes(AddrNMICopyBuf, NMIoffs, NMICopyBuf);

    static const unsigned char VScrollSpeed[] =
    {
        0xA5,0x33,      //   lda $33
        0x29,0x03,      //   and #$03
        0xD0,6,         //   bne +
        0x20,0xC4,0xC6, //   jsr LoadEnemyGraphics
        
        0x20, AddrNMICopyBuf&255, AddrNMICopyBuf>>8, // jsr AddrNMICopyBuf
     //   0x20,0x1B,0xC0, //   jsr NextFrame
        0x60            // + rts 
    };
    unsigned FixVScrollSpeed = ROM.FindFreeSpaceIn(5, sizeof(VScrollSpeed));
    ROM.WriteBytes(FixVScrollSpeed, sizeof(VScrollSpeed), VScrollSpeed);
    
    ROM.Write(0x05A088, 0x20); // jsr
    ROM.Write(0x05A089, FixVScrollSpeed&0xFF); //
    ROM.Write(0x05A08A, FixVScrollSpeed >> 8); // addr
    NopRegion(0x05A08B, 3);
#endif
}
